import numpy as np

import imglyb
from .accesses import *
from .types import for_np_dtype
from jnius import autoclass, PythonJavaClass, java_method, cast, JavaException

Cell                        = autoclass('net.imglib2.img.cell.Cell')
CellGrid                    = autoclass('net.imglib2.img.cell.CellGrid')
SoftRefLoaderCache          = autoclass('net.imglib2.cache.ref.SoftRefLoaderCache')
CachedCellImg               = autoclass('net.imglib2.cache.img.CachedCellImg')
GuardedStrongRefLoaderCache = autoclass('net.imglib2.cache.ref.GuardedStrongRefLoaderCache')
PythonHelpers               = autoclass('net.imglib2.python.Helpers')
VolatileViews               = autoclass('bdv.util.volatiles.VolatileViews')
VolatileByteArray           = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileByteArray')
Random                      = autoclass('java.util.Random')

def as_cached_cell_img(func, cell_grid, dtype, cache_generator=SoftRefLoaderCache, volatile_access=False):
    loader      = CacheLoaderFromFunction(func, cell_grid, volatile_access)
    cache       = cache_generator().withLoader(loader)
    t           = for_np_dtype(dtype, volatile=False)
    img         = CachedCellImg(cell_grid, t.getEntitiesPerPixel(), cache, as_array_access(np.zeros((0,), dtype=dtype), volatile=volatile_access))
    linked_type = t.getNativeTypeFactory().createLinkedType(img)
    img.setLinkedType(linked_type)
    img.randomAccess()
    return img

def wrap_volatile(cell_img, dirty=False):
    return PythonHelpers.createVolatileCachedCellImg(cell_img, dirty)

class MakeAccess(PythonJavaClass):
    __javainterfaces__ = ['java/util/function/Function']

    def __init__(self, func):
        super(MakeAccess, self).__init__()
        self.func = func

    @java_method('(Ljava/lang/Object;)Ljava/lang/Object;')
    def apply(self, t):
        return self.func(t)


class CacheLoaderFromFunction(PythonJavaClass):
    __javainterfaces__ = ['net/imglib2/cache/CacheLoader']

    def __init__(self, func, cell_grid, volatile=False):
        super(CacheLoaderFromFunction, self).__init__()
        self.func      = func
        self.cell_grid = cell_grid
        self.volatile  = volatile

    @java_method('(Ljava/lang/Object;)Ljava/lang/Object;', name='get')
    def get(self, index):
        chunk      = self.func(index)
        refGuard   = imglyb.util.ReferenceGuard(chunk)
        address    = chunk.ctypes.data

        try:
            pos    = PythonHelpers.cellMin(self.cell_grid, index)
            target = as_array_access(chunk, volatile=self.volatile)
            cell   = Cell(chunk.shape[::-1], pos, target)
        except JavaException as e:
            print('Name        --', e.classname)
            print('Message     --', e.innermessage)
            print('Stack trace --', e.stacktrace)
            if e.stacktrace:
                for line in e.stacktrace:
                    print(line)
            raise e

        return cast('net.imglib2.img.cell.Cell', cell)

try:
    import dask
    import dask.array
    def dask_array_as_cached_cell_img(
            dask_array,
            chunk_size=None,
            cache_generator=SoftRefLoaderCache,
            volatile_access=False):
        if not dask.array.core._check_regular_chunks(dask_array.chunks):
            raise ValueError('Expected dask array with regular chunking but got {}'.format(dask_array.chunks))
        slices     = dask.array.core.slices_from_chunks(dask_array.chunks)
        dims       = dask_array.shape
        block_size = chunk_size if chunk_size else tuple(c[0] for c in dask_array.chunks)
        return as_cached_cell_img(
            func            = lambda index : dask_array[slices[index]].compute(),
            cell_grid       = PythonHelpers.makeGrid(dims, block_size),
            dtype           = dask_array.dtype,
            cache_generator = cache_generator,
            volatile_access  = volatile_access
        )

except ImportError as e:
    print("[WARN] Unable to import dask -- dask to imglib2 wrappers not available!")



import numpy as np

import imglyb
from jnius import autoclass


Accesses            = autoclass('tmp.net.imglib2.img.basictypeaccess.Accesses')

ByteArray           = autoclass('net.imglib2.img.basictypeaccess.array.ByteArray')
CharArray           = autoclass('net.imglib2.img.basictypeaccess.array.CharArray')
DoubleArray         = autoclass('net.imglib2.img.basictypeaccess.array.DoubleArray')
FloatArray          = autoclass('net.imglib2.img.basictypeaccess.array.FloatArray')
IntArray            = autoclass('net.imglib2.img.basictypeaccess.array.IntArray')
LongArray           = autoclass('net.imglib2.img.basictypeaccess.array.LongArray')
ShortArray          = autoclass('net.imglib2.img.basictypeaccess.array.ShortArray')

VolatileByteArray   = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileByteArray')
VolatileCharArray   = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileCharArray')
VolatileDoubleArray = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileDoubleArray')
VolatileFloatArray  = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileFloatArray')
VolatileIntArray    = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileIntArray')
VolatileLongArray   = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileLongArray')
VolatileShortArray  = autoclass('net.imglib2.img.basictypeaccess.volatiles.array.VolatileShortArray')

ByteUnsafe          = autoclass('net.imglib2.img.basictypelongaccess.unsafe.ByteUnsafe')
CharUnsafe          = autoclass('net.imglib2.img.basictypelongaccess.unsafe.CharUnsafe')
DoubleUnsafe        = autoclass('net.imglib2.img.basictypelongaccess.unsafe.DoubleUnsafe')
FloatUnsafe         = autoclass('net.imglib2.img.basictypelongaccess.unsafe.FloatUnsafe')
IntUnsafe           = autoclass('net.imglib2.img.basictypelongaccess.unsafe.IntUnsafe')
LongUnsafe          = autoclass('net.imglib2.img.basictypelongaccess.unsafe.LongUnsafe')
ShortUnsafe         = autoclass('net.imglib2.img.basictypelongaccess.unsafe.ShortUnsafe')

# does not work with strided accesses, currently
def as_array_access(ndarray, volatile=False):
    if ndarray.dtype == np.uint8 or ndarray.dtype == np.int8:
        return _as_array_access(ndarray, ByteUnsafe, lambda n : VolatileByteArray(n, True) if volatile else ByteArray(n))
    elif ndarray.dtype == np.uint16 or ndarray.dtype == np.int16:
        return _as_array_access(ndarray, ShortUnsafe, lambda n : VolatileShortArray(n, True) if volatile else ShortArray(n))
    elif ndarray.dtype == np.uint32 or ndarray.dtype == np.int32:
        return _as_array_access(ndarray, IntUnsafe, lambda n : VolatileIntArray(n, True) if volatile else IntArray(n))
    elif ndarray.dtype == np.uint64 or ndarray.dtype == np.int64:
        return _as_array_access(ndarray, LongUnsafe, lambda n : VolatileLongArray(n, True) if volatile else LongArray(n))
    elif ndarray.dtype == np.float32:
        return _as_array_access(ndarray, FloatUnsafe, lambda n : VolatileFloatArray(n, True) if volatile else FloatArray(n))
    elif ndarray.dtype == np.float64:
        return _as_array_access(ndarray, DoubleUnsafe, lambda n : VolatileDoubleArray(n, True) if volatile else DoubleArray(n))

def _as_array_access(ndarray, src_clazz, tgt_clazz):
    src = src_clazz(ndarray.ctypes.data)
    tgt = tgt_clazz(ndarray.size)
    Accesses.copyAny(src, 0, tgt, 0, ndarray.size)
    return tgt




"""

This script is a wrapper to allow imglib2-imglyb code that uses Java's AWT
to run properly on OS X.  It starts the Cocoa event loop before Java and
keeps Cocoa happy  See https://github.com/kivy/pyjnius/issues/151 for more.

In particular, this wrapper allows one to run the code from imglyb-examples.

Tested in Python 3 only!  Not sure if you'd need to change anything for Python 2
beyond the print statements.

usage: python OSXAWTwrapper.py [module name | script path] [module or script parameters]

"""

import os
import sys


import objc
from Foundation import *
from AppKit import *
from PyObjCTools import AppHelper

usage = "usage: python OSXAWTwrapper.py [module name | script path] [module or script parameters]"

def runAwtStuff():

    import runpy

    # user can provide either a module or a path to a script;
    #   either way, need to remove it from sys.argv,
    #   because the wrapped module or script might parse sys.argv for its own reasons:
    if len(sys.argv) > 1:
        name = sys.argv[1]
        sys.argv.remove(name)

        # whether module or script, need to set the run_name for things to work as expected!
        try:
            if os.path.exists(name):
                runpy.run_path(name, run_name="__main__")
            else:            
                runpy.run_module(name, run_name="__main__")
        except Exception as e:
            print("exception occurred while running {}: {}".format(name, e))

            ## lots can go wrong here, and exceptions can bubble up from
            #   the Java layer, too; uncomment lines below to print
            #   more information on exception
            # note: different exceptions have different attributes, so you
            #   might need to adjust the lines below; use dir(e) to see
            #   what you have available when you are debugging
            # print("exception details: ")
            # print("e.args: ", e.args)
            # print("e.__class__: ", e.__class__)
            # print("e.stacktrace: ")
            # for line in e.stacktrace:
            #     print("\t", line)
            ## if Java throws a reflection error, you might want this:
            ## print("e.innermessage", e.innermessage)
    else:
        print(usage)
        print("no module or script specified")


class AppDelegate (NSObject):
    def init(self):
        self = objc.super(AppDelegate, self).init()
        if self is None:
            return None
        return self

    def runjava_(self, arg):
        runAwtStuff()
        # we need to terminate explicitly, or it'll hang when
        #   the wrapped code exits
        NSApp().terminate_(self)

    def applicationDidFinishLaunching_(self, aNotification):
        self.performSelectorInBackground_withObject_("runjava:", 0)


def main():
    app = NSApplication.sharedApplication()
    delegate = AppDelegate.alloc().init()
    NSApp().setDelegate_(delegate)
    # this is necessary to have keyboard events sent to the UI;
    #   basically this call makes the script act like an OS X application,
    #   with Dock icon and everything
    NSApp.setActivationPolicy_(NSApplicationActivationPolicyRegular)
    AppHelper.runEventLoop()

if __name__ == '__main__' : 
    main()




import os

__all__ = ('set_imglib2_imglyb_version', 'get_imglib2_imglyb_version')

version = '0.3.4'

_default_imglib2_imglyb_version = '0.3.0'
_imglib2_imglyb_version         = os.getenv('IMGLIB2_IMGLYB_VERSION', _default_imglib2_imglyb_version)

def set_imglib2_imglyb_version(version):
    global _imglib2_imglyb_version
    _imglib2_imglyb_version = version

def get_imglib2_imglyb_version():
    global _imglib2_imglyb_version
    return _imglib2_imglyb_version
